local L = Capitalist.Locale
local realm = GetRealmName()
local character = UnitName("player")
local faction = UnitFactionGroup("player")

local AceConfigRegistry = LibStub("AceConfigRegistry-3.0")

function Capitalist:EnsureNoCostIsSet(productName)
	local cost = self.db.profile.components[productName].cost
	if (cost) then
		local costString = self:PrettyCurrency(cost)
		self:Print(string.format("  "..L["Removing cost (%s) from product (%s)."], costString, self:GetComponentText(productName)))
		self.db.profile.components[productName].cost = nil
	end

	-- REVIEW: In light of Enchanting Vellum now being merchant-supplied, does the assumption of the following logic still hold?
	local buyPrice = self.db.profile.components[productName].buyPrice
	if (buyPrice) then
		local buyPriceString = self:PrettyCurrency(buyPrice)
		self:Print(string.format("  "..L["Removing buy price (%s) from product (%s)."], buyPriceString, self:GetComponentText(productName)))
		self.db.profile.components[productName].buyPrice = nil
	end
end

-- products are not allowed to have costs set directly
-- update data in case products were discovered after
-- costs were previously set on the product
-- (i.e. it was treated as a resource)
function Capitalist:EnforceProductsCannotHaveCostSetting()
	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			for skillName, skillData in pairs(professionData) do
				local productName = self:ProductNameFromSkillName(profession, skillName)
				if (not self:SkillCooldownApplies(profession, skillName) and
					self:IsComponent(productName)) then

					self:EnsureNoCostIsSet(productName)
				end
			end
		end
	end

	if (self.db.profile.enableSalvageSupply) then
		local oRealm = self:GetProfile_Realm_Faction(realm, faction)
		for crafter, characterData in pairs(oRealm) do
			for characterSection, characterSectionData in pairs(characterData) do
				if (not self.db.profile.professions[characterSection]) then
					-- Assume non-profession sections are salvage skill sections
					local salvageSkill = characterSection
					local salvageSkillData = characterSectionData
					local salvageSkillLink = characterSectionData.recipeLink

					for skillTarget, skillTargetData in pairs(characterSectionData) do
						if ((skillTarget ~= "recipeLink") and
							(skillTarget ~= "numConsume")) then
							-- Assume non-recipeLink, non-numConsume entries are skillTarget sections

							local skillTargetName = self:LinkExtractName(skillTarget)

							if (skillTargetName) then
								if (self:IsProduct(skillTargetName) or self:IsResource(skillTargetName)) then
									for skillTargetResource, skillTargetResourceCount in pairs(skillTargetData.resourceTally) do
										local skillTargetResourceName = self:LinkExtractName(skillTargetResource)

										-- TODO: 1 - add logic to build prices for salvaging
										-- TODO: 2 - enable the function call
--										print("TODO - EnsureNoCostIsSet: "..self:LinkExtractName(skillTargetResource))
										--self:EnsureNoCostIsSet(skillTargetResourceName)
									end
--									print("TODO - Cause(+): "..skillTarget)
								else
									--print("TODO - Cause(-): "..skillTarget)
								end
							end
						end
					end					
				end
			end
		end
	end
end

function Capitalist:ApplyFilters()
	self:DebugEntry("ApplyFilters")

	-- reset memory
	self:ResourceMemoryReset()
	self:PartMemoryReset()
	self:ProductMemoryReset()

	-- CONSIDER:  Automatic queueing of parts that alts need to make

	local productsToProcess = {}
	local resourcesToMakeVisible = {}

	local q = 0

	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			productsToProcess[profession] = { skills = {}, pending = 1, }

			self:DebugMisc(profession)

			for skillName, skillData in pairs(professionData) do
				-- for products not allowed, mark them as blocked
				local productBlocked
				if (not self:ProductIsAllowed(profession, skillName)) then
					productBlocked = true
				end

				-- for products with unavailable product name, block as well
				local productName = self:ProductNameFromSkillName(profession, skillName)
				if (not productName) then
					productBlocked = true
				end

				self:ProductSetVisible(profession, skillName, not productBlocked)
				if (not productBlocked) then
					if (self:ProductIsEnabled(profession, skillName)) then
						table.insert(productsToProcess[profession].skills, skillName)
					end
				end
			end
		end
	end

	local doneWithProducts
	while (not doneWithProducts) do
		doneWithProducts = true
		for profession, professionData in pairs(self.db.profile.professions) do
			if (productsToProcess[profession].skills) then
				while (productsToProcess[profession].pending <= #productsToProcess[profession].skills) do
					local skillName = productsToProcess[profession].skills[productsToProcess[profession].pending]

					local includeSkillForShop = true
					if (not self.db.profile.optionsSpecific[realm][character].crossShop) then
						if (not self:GetProfile_Realm_Faction_Character_Profession_Skill(realm, faction, character, profession, skillName)) then
							includeSkillForShop = false
						end														
					end

					if (includeSkillForShop) then
						local product = self.db.profile.professions[profession][skillName]

						local components = self:ComponentsForSkillProduct(profession, skillName)
						for componentName, componentCount in pairs(components) do
							local componentSkillName, componentProfession = self:SkillNameFromProductName(componentName)
							if (componentSkillName and componentProfession) then
								self:PartSetVisible(componentProfession, componentName, true)
								if (not tContains(productsToProcess[componentProfession].skills, componentSkillName)) then
									self:ProductSetVisible(componentProfession, componentSkillName, true)
									table.insert(productsToProcess[componentProfession].skills, componentSkillName)
									doneWithProducts = false
								end
							else
								--if (not tContains(resourcesToMakeVisible, componentName)) then
									--table.insert(resourcesToMakeVisible, componentName)
								--end
								if (not resourcesToMakeVisible[componentName]) then
									resourcesToMakeVisible[componentName] = { }
									resourcesToMakeVisible[componentName].professions = { }
								end
								if (not tContains(resourcesToMakeVisible[componentName].professions, profession)) then
									table.insert(resourcesToMakeVisible[componentName].professions, profession)
								end
							end
						end
					end

					productsToProcess[profession].pending = productsToProcess[profession].pending + 1
				end
			end
		end
	end

	for resourceName, resourceData in pairs(resourcesToMakeVisible) do
		-- make visible resources that are in use
		if (self:ResourceGetMerchant(resourceName)) then
			-- resources with merchants suppliers don't have
			-- a separate cost value; they use the buyPrice
			if (self:HasResourceCopperCost(resourceName)) then
				local cost = self:GetResourceCopperCost(resourceName)
				local costString = self:PrettyCurrency(cost)
				self:Print(string.format("  "..L["Removing cost (%s) from merchant-supplied resource (%s)."], costString, self:GetComponentText(resourceName)))
				self:SetResourceCopperCost(resourceName, nil)
			end
		else
			for i=1, #(resourceData.professions) do
				self:ResourceSetVisible(resourceData.professions[i], resourceName, true)
			end
		end
	end

	self:BuildList()
	self:UpdateProductOptions()
	self:UpdateResourceCosts()
	self:UpdateStatistics()

	self:DebugExit("ApplyFilters")
end

-- FEATURE: DisallowEnchantingVellumRecipe
function Capitalist:IsExcludedSkill(profession, skillName)
	if ((profession == L["Inscription"]) and
		(skillName == L["Enchanting Vellum"])) then

		return true
	end

	return false
end

-- scans the trade skill frame and adds all products the player knows
function Capitalist:ScanTradeSkillFrame(...)
	self:DebugEntry("ScanTradeSkillFrame")

	local profession = ...;
	self:DebugMisc(profession)

	if (profession == L["Runeforging"]) then
		return self:DebugExit("ScanTradeSkillFrame")
	end

	-- REVIEW: Update products for other players, and maybe track their names to do business later
	-- Bypass data updating when viewing another player's profession information
	if (IsTradeSkillLinked()) then
		return self:DebugExit("ScanTradeSkillFrame")
	end

	SetTradeSkillCategoryFilter(0)
	SetTradeSkillInvSlotFilter(0, 1, 1)
	SetTradeSkillItemNameFilter("")

	-- REVIEW: Workaround for cache not full ready yet
	local _, skillTypeHeader, _, _, _ = GetTradeSkillInfo(1)
	if (skillTypeHeader ~= "header") then
		return self:DebugExit("ScanTradeSkillFrame")
	end
	
	self:EnsureProfile_Realm_Faction_Character_Profession(realm, faction, character, profession)

	local newProducts = 0
	local newComponents = 0
	local newCharacterProducts = 0

	if (not self.db.profile.professions) then
		self.db.profile.professions = {}
	end
	if (not self.db.profile.professions[profession]) then
		self.db.profile.professions[profession] = {}
	end
	
	if (not self.db.profile.components) then
		self.db.profile.components = {}
	end
	
	local sellPriceUpdated
	for skillId=1, GetNumTradeSkills() do

		local skillName, skillType, numAvailable, isExpanded, altVerb = GetTradeSkillInfo(skillId)
		self:DebugMisc(skillName)		-- "Enchant Bracer - Major Stamina"
		self:DebugMisc(skillType)		-- "header" or one of ("trivial", "easy", "medium", "optimal", "difficult")
		self:DebugMisc(numAvailable)	-- the bracketed value shown in the craft UI, showing how many you can make with current components
		self:DebugMisc(isExpanded)		--
		self:DebugMisc(altVerb)			-- nil or "Enchant" (REVIEW: perhaps others for armor kits?)

		if (not self:IsExcludedSkill(profession, skillName)) then
			local itemLink = GetTradeSkillItemLink(skillId)
			self:DebugMisc(itemLink)

			local recipeLink = GetTradeSkillRecipeLink(skillId)
			self:DebugMisc(recipeLink)

			if (itemLink) then
				skillName = self:ApplyGlyphProductNameCorrection(skillName, itemLink)
				skillName = self:ApplyCraftedMalevolentNameCorrection(skillName, itemLink)
			end

			if ((skillType ~= "header") and L.checkForValidSkillName(profession, skillName)) then
				local skillDescription = GetTradeSkillDescription(skillId)
				self:DebugMisc(skillDescription)

				local minMade = 1
				local maxMade = 1	
				if (profession ~= "Enchanting") then
					minMade, maxMade = GetTradeSkillNumMade(skillId)
				end
				self:DebugMisc(minMade)
				self:DebugMisc(maxMade)

				local _, cooldownApplies = GetTradeSkillCooldown(skillId)
				if (not cooldownApplies) then
					cooldownApplies = nil
				end

				if (maxMade == minMade) then
					maxMade = nil
					if (minMade == 1) then
						minMade = nil
					end
				end

				-- add it to the known products
				if (not self.db.profile.professions[profession][skillName]) then
					self.db.profile.professions[profession][skillName] = {
																			components = {},
																		 }

					newProducts = newProducts + 1
				end

				-- add the known products for this character
				if (not self:GetProfile_Realm_Faction_Character_Profession_Skill(realm, faction, character, profession, skillName)) then
					self:EnsureProfile_Realm_Faction_Character_Profession_Skill(realm, faction, character, profession, skillName)
					newCharacterProducts = newCharacterProducts + 1
				end


				-- CONSIDER: For all information extractable from link, itemLink, or recipeLink *don't store it*.

				if (self.db.profile.professions[profession][skillName].description ~= skillDescription) then
					self.db.profile.professions[profession][skillName].description = skillDescription
				end

				if (self.db.profile.professions[profession][skillName].itemLink ~= itemLink) then
					self.db.profile.professions[profession][skillName].itemLink = itemLink
				end

				if (self.db.profile.professions[profession][skillName].recipeLink ~= recipeLink) then
					self.db.profile.professions[profession][skillName].recipeLink = recipeLink
				end

				if (self.db.profile.professions[profession][skillName].min ~= minMade) then
					self.db.profile.professions[profession][skillName].min = minMade
				end
				if (self.db.profile.professions[profession][skillName].max ~= maxMade) then
					self.db.profile.professions[profession][skillName].max = maxMade
				end

				if (self.db.profile.professions[profession][skillName].cooldownApplies ~= cooldownApplies) then
					self.db.profile.professions[profession][skillName].cooldownApplies = cooldownApplies
					print("cooldown @ "..skillName)
				end

				local oSkillName = self:GetProfile_Realm_Faction_Character_Profession_Skill(realm, faction, character, profession, skillName)
				if (oSkillName.type ~= skillType) then
					oSkillName.type = skillType
				end

				if (itemLink) then
					local _, _, _, _, _, _, _, _, _, _, itemSellPrice = GetItemInfo(itemLink)
					if (itemSellPrice and (itemSellPrice > 0)) then
						if (self.db.profile.professions[profession][skillName].sellPrice ~= itemSellPrice) then
							self.db.profile.professions[profession][skillName].sellPrice = itemSellPrice
							self:ClearValuesWithValueSrc(skillName)
							sellPriceUpdated = true
						end
					end
				end

				local componentsScanned = {}
				local componentsComplete = false

				for reagentId=1, GetTradeSkillNumReagents(skillId) do
					local reagentName, reagentTexture, reagentCount, playerReagentCount = GetTradeSkillReagentInfo(skillId, reagentId)
					self:DebugMisc(reagentName)
					self:DebugMisc(reagentTexture)
					self:DebugMisc(reagentCount)
					self:DebugMisc(playerReagentCount)

					if (reagentName and (reagentCount > 0)) then
						componentsComplete = true
						componentsScanned[reagentName] = reagentCount

						-- for previously unknown components, save the item link and the id
						if (not self.db.profile.components[reagentName]) then
							self:Print("Undiscovered reagent: "..reagentName)
							local reagentLink = GetTradeSkillReagentItemLink(skillId, reagentId)
							self:DebugMisc(reagentLink)

							if (reagentLink) then
								-- TODO: Optimization - don't store this ID
								local _, _, reagentItemId = reagentLink:find("Hitem:(%d+):")
								self:DebugMisc(reagentItemId)
					
								self.db.profile.components[reagentName] = {
																		link = reagentLink,
																		id = tonumber(reagentItemId),
																	  }

								newComponents = newComponents + 1
							end
						end
					else
						componentsComplete = false
						break
					end
				end

				if (componentsComplete) then
					local dbComponents = self.db.profile.professions[profession][skillName].components

					for oldComponentName, oldComponentCount in pairs(dbComponents) do
						local newComponentCount = componentsScanned[oldComponentName]
						if (not newComponentCount) then
							dbComponents[oldComponentName] = nil
						elseif (newComponentCount ~= oldComponentCount) then
							dbComponents[oldComponentName] = componentsScanned[oldComponentName]
							componentsScanned[oldComponentName] = nil
						else
							componentsScanned[oldComponentName] = nil
						end
					end

					for newComponentName, newComponentCount in pairs(componentsScanned) do
						dbComponents[newComponentName] = newComponentCount
					end
				end
			end
		end
	end

	-- REVIEW: Still needed?  If so, make smarter about only apply to enchants.
	if (profession == L["Enchanting"]) then
		local componentName = L["Enchanting Vellum"]
		if (not self.db.profile.components[componentName]) then
			local itemLink = select(2, GetItemInfo(self:GetComponentId("Enchanting Vellum")))
			self.db.profile.components[componentName] = {
															id = self:GetComponentId(componentName),
															link = itemLink,
														}
			newComponents = newComponents + 1
		end
	end

	if (sellPriceUpdated or
		(newProducts + newComponents + newCharacterProducts > 0)) then
		self:Print(string.format(
						L["Found %d new products and %d new components."],
						newProducts,
						newComponents
					))
		self:Print(string.format(
						L["Found %d new character products."],
						newCharacterProducts
					))

		self:EnforceProductsCannotHaveCostSetting()
		self:UpdateComponentValues()
		self:UpdateCombinedValues()
	
		--TODO: remove after debugging
--		self:EnforceProductsCannotHaveCostSetting()

		self:ResetSkillCaches()
		self:ApplyFilters()
		self:UpdateProductOptions()
		self:UpdateResourceCosts()
		self:UpdateStatistics()
		AceConfigRegistry:NotifyChange("Capitalist")
	end
	
	self:UpdateRestockStack(profession)

	self:DebugExit("ScanTradeSkillFrame")
end

function Capitalist:BuildList()
	self:DebugEntry("BuildList")

	self.list = {}

	-- Generally, it's "simple bucketing" -- 1 full product name per bucket
	self.scanBuckets = {}
	-- Optimization 1: do "Glyph of " in batch
	local productsIncludeGlyph = false
	-- Optimization 2: do "Scroll of Enchant " and "Enchant " via "Enchant " batch
	local productsIncludeScroll = false

	local gemBuckets = {}

	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			self:DebugMisc(profession)

			for skillName, skillData in pairs(professionData) do

				local includeSkillForScan = true
				if (not self.db.profile.optionsSpecific[realm][character].crossScan) then
					if (not self:GetProfile_Realm_Faction_Character_Profession_Skill(realm, faction, character, profession, skillName)) then
						includeSkillForScan = false
					end														
				end

				if (includeSkillForScan) then
					local productName = self:ProductNameFromSkillName(profession, skillName)
					if (productName and
						self:ProductIsVisible(profession, skillName) and
						self:ProductIsEnabled(profession, skillName)) then

						self.list[productName] = { profession = profession, skillName = skillName }

						if (self:ProductIsEnabled(profession, skillName)) then
							local applyOptimizationGem = false
							local applyOptimizationGlyph = false
							local applyOptimizationScroll = false

							if (profession == L["Jewelcrafting"]) then
								local gem = string.match(productName, "[%a]+ (.*)")
								if (gem) then
									if (not tContains(gemBuckets, gem)) then
										table.insert(gemBuckets, gem)
									end
									applyOptimizationGem = true
								end
							end

							if (productName:find(L["Glyph of "])) then
								applyOptimizationGlyph = true
							end
							if (productName:find(L["Enchant "])) then
								applyOptimizationScroll = true
							end

							if (applyOptimizationGem) then
								-- nop, handled via gemBuckets
							elseif (applyOptimizationGlyph) then
								productsIncludeGlyph = true
							elseif (applyOptimizationScroll) then
								productsIncludeScroll = true
							else
								table.insert(self.scanBuckets, productName)
							end
						end
					end
				end
			end
		end
	end

	for i=1, #gemBuckets do
		local gem = gemBuckets[i]
		table.insert(self.scanBuckets, 1, " "..gem)
	end

	if (productsIncludeGlyph) then
		table.insert(self.scanBuckets, 1, L["Glyph of "])
	end
	if (productsIncludeScroll) then
		table.insert(self.scanBuckets, 1, L["Enchant "])
	end

	-- resolve overlapping buckets
	local i=1
	local j=1
	while (i <= #self.scanBuckets) do
		while (j <= #self.scanBuckets) do
			if ((not (i == j)) and
				(self.scanBuckets[j]:find(self.scanBuckets[i]))) then
				table.remove(self.scanBuckets, j)
				if (j < i) then
					i = i - 1
				end
			else
				j = j + 1
			end
		end

		i = i + 1
		j = 1
	end

	self:DebugExit("BuildList")
end
